#!/usr/bin/env python
"""
Providing general utilities for working with different formats of molecular files
"""
import os
import glob
import time
from collections import defaultdict, OrderedDict
from tqdm import tqdm
import numpy as np
from rdkit import Chem
from rdkit.Chem import rdMolAlign, inchi
from rdkit.Chem.rdMolDescriptors import CalcNumUnspecifiedAtomStereoCenters
from typing import List, Tuple, Dict, Union, Optional, Callable
from collections import defaultdict

def guess_file_type(filename):
    """Returns the extension for the filename"""
    assert '.' in filename
    return os.path.splitext(filename)[1][1:]

# Functions related to smi files
def smiles2smi(smiles: List[str], path: str) -> str:
    """Converting a list of smiles into a smi file,
    naming each SMILES using inchikey"""
    lines = []
    for smi in smiles:
        mol = Chem.MolFromSmiles(smi)
        inchikey = inchi.MolToInchiKey(mol)
        lines.append(f"{smi}  {inchikey}\n")
    with open(path, "w+") as f:
        for line in lines:
            f.write(line)
    return path

def report(path: str):
    """Given a smi file, reports the following:
    - number of SMILES
    - SMILES size distribution
    - element type and percent of molecules with this types
    - number of charged molecules
    - number of SMILES with unspecified stereo center
    """
    suppl = Chem.SmilesMolSupplier(path, titleLine=False)
    c = 0  #count number of SMILES
    sizes = []
    element_counts = defaultdict(lambda: 0)
    charges = []
    charge_counts = defaultdict(lambda: 0)
    num_charged_mols = 0
    unspecified_atom_centers = []
    num_unspecified_mols = 0
    for mol in suppl:
        c += 1
        atoms = [a.GetAtomicNum() for a in mol.GetAtoms()]  #H not included
        elements = list(set(atoms))
        sizes.append(len(atoms))
        for e in elements:
            element_counts[e] += 1
        
        charge = Chem.rdmolops.GetFormalCharge(mol)
        charge_counts[charge] += 1
        charges.append(charge)
        if charge != 0:
            num_charged_mols += 1
        
        unspecified_centers = CalcNumUnspecifiedAtomStereoCenters(mol)
        unspecified_atom_centers.append(unspecified_centers)
        if unspecified_centers > 0:
            num_unspecified_mols += 1
    
    print("Total number of SMILES: ", c, flush=True)
    print(f"SMILES size distribution: mean={str(np.mean(sizes))} std={str(np.std(sizes))} min={str(min(sizes))} max={str(max(sizes))}", flush=True)    
    print("Breakdown of element types and its prevailance: ", flush=True)
    for e, c_e in sorted(element_counts.items()):
        print(f"    {str(e)} total: {str(c_e)}  percent: {str(round(c_e/c, 3))}", flush=True)
    print(f"Number of charged molecules: {str(num_charged_mols)}", flush=True)
    print("Breakdown of charge distribution", flush=True)
    for charge, c_c in sorted(charge_counts.items()):
        print(f"    charge={str(charge)} total: {str(c_c)} percent: {str(round(c_c/c, 3))}", flush=True)
    print(f"Number of molecules with unspecified atomic centers: {str(num_unspecified_mols)}", flush=True)

def combine_smi(smies: List[str], out: str):
    """Combine smi files into a single file"""
    data = []
    for smi in smies:
        with open(smi, 'r') as f:
            datai = f.readlines()
        data += datai
    data = list(set(data))
    with open(out, 'w+') as f2:
        for line in data:
            if not line.isspace():
                f2.write((line.strip() + '\n'))

def is_macrocycle(smiles:str, size=10):
    """Check if a SMIELS contains a macrocycle part (a 10-membered or large 
    ring regardless of their aromaticity and hetero atoms content)"""
    mol = Chem.MolFromSmiles(smiles)
    ring = mol.GetRingInfo()
    ring_bonds = ring.BondRings()
    for bonds in ring_bonds:
        if len(bonds) >= size:
            return True
    return False

def split_smi(smi):
    """Split an input .smi file into two files:
    one contains small SMILES, the other contain macrocycle smiles
    """
    #Prepare out file path
    dir = os.path.dirname(os.path.realpath(smi))
    basename = os.path.basename(smi)
    normal_name = basename.split(".")[0].strip() + "_normal.smi"
    macrocycle_name = basename.split(".")[0].strip() + "_macrocycle.smi"
    normal_path = os.path.join(dir, normal_name)
    macrocycle_path = os.path.join(dir, macrocycle_name)

    normal = []
    macrocycle = []
    with open(smi, "r") as f:
        data = f.readlines()
    for line in tqdm(data):
        smi_idx = line.strip().split()
        smi = smi_idx[0]
        if is_macrocycle(smi):
            macrocycle.append(line)
        else:
            normal.append(line)
    
    l_ = len(normal) + len(macrocycle)
    l = len(data)
    assert(l == l_)

    with open(normal_path, "w+") as f:
        for line in normal:
            f.write(line)

    with open(macrocycle_path, "w+") as f:
        for line in macrocycle:
            f.write(line)

# Functions related to SDF files
def countSDF(sdf):
    """Counting the number of structures in SDF file"""
    mols = Chem.SDMolSupplier(sdf)
    mols2 = [mol for mol in mols]
    c = len(mols2)
    return c

def SDF2chunks(sdf:str)->List[List[str]]:
    """given a sdf file, return a list of chunks,
    each chunk consists of lines of a molecule as they appear in the original file"""
    chunks = []
    with open(sdf, "r") as f:
        data = f.readlines()
    chunk = []
    for line in data:
        if line.strip() == "$$$$":
            chunk.append(line)
            chunks.append(chunk)
            chunk = []
        else:
            chunk.append(line)
    return chunks
       
# Functions related to XYZ files
def combine_xyz(in_folder, out_path):
    """
    Combining all xyz files in the in_folder into a single xyz file (out_path).

    Arguemnts:
        in_folder: a folder contains all xyz files.
        out_path: a path of xyz file to store every structure in the in_folder

    Returns:
        Combining all xyz files in the in_folder into out_path.
    """
    file_paths = os.path.join(f"{in_folder}/*.xyz")
    files = glob.glob(file_paths)
    # print(f'There are {len(files)} single xyz files...')

    results = []
    for file in files:
        with open(file, 'r') as f:
            data = f.readlines()
        assert(len(data) == (int(data[0]) + 2))
        results += data

    with open(out_path, 'w+') as f:
        for line in results:
            f.write(line)
    # print(f'Combined in a singl file {out_path}!')

# Functions involving multiple molecular file formats
def to_smiles(path, fomat="sdf"):
    """converting a file from a given format to smi file
    input: path
    format: [optional] sdf
    
    returns: a path of smi file containing the same molecules as in the sdf"""
    suppl = Chem.SDMolSupplier(path)
    smiles = []
    for i, mol in enumerate(suppl):
        name = mol.GetProp("_Name").strip()
        if len(name) == 0:  #len("") == 0
            name = str(i)
        smiles.append((Chem.MolToSmiles(mol, isomericSmiles=True, canonical=True), name))
    
    #write
    folder = os.path.dirname(os.path.abspath(path))
    new_base = os.path.basename(path).split(".")[0].strip() + ".smi"
    new_path = os.path.join(folder, new_base)
    with open(new_path, "w") as f:
        for smi, name in smiles:
            f.write(f"{smi} {name}\n")
    return new_path

def find_smiles_not_in_sdf(smi, sdf):
    """Find the SMILES who doesn't have a 3D structure in the SDF file
    smi: path to an smi file (the input path for Auto3D)
    sdf: path to an SDF file"""
    #find all SMILES ids
    smi_names = []
    with open(smi, "r") as f:
        data = f.readlines()
    for line in data:
        smi, id = tuple(line.strip().split())
        smi_names.append((smi.strip(), id.strip()))
    
    sdf_data = []
    mols = Chem.SDMolSupplier(sdf)
    for mol in mols:
        sdf_data.append(mol.GetProp("_Name"))
    sdf_data = list(set(sdf_data))

    bad = []
    for smi, id in smi_names:
        has_3D_structure = False
        # for line in sdf_data:
        #     if id in line:
        #         has_3D_structure = True
        if id in sdf_data:
            has_3D_structure = True
        if not has_3D_structure:
            bad.append((id, smi))

    if len(bad) > 0:
        print("The following SMILES has no 3D structure in the SDF file.", flush=True)
        print("ID, SMILES", flush=True)
        for id, smi in bad:
            print(id, smi, flush=True)
    else:
        print("Every SMILES has at least an 3D structure in the SDF file.", flush=True)
    return bad

def encode_ids(path: str) -> Tuple[str, dict]:
    '''For a smi/SDF Files, encode the ids into numbers,
    return the new smi files path and a dictionary containing the mapping'''
    basename = os.path.basename(path)
    dir = os.path.dirname(os.path.abspath(path))
    extension = basename.split('.')[-1].strip()
    new_path = os.path.join(dir, basename.split('.')[0].strip() + '_encoded.' + extension)

    if extension == 'smi':
        new_data = []
        with open(path, 'r') as f:
            data = f.readlines()
        mapping = {}
        for i, line in enumerate(data):
            if line.isspace():
                continue
            smi, id = line.strip().split()
            mapping[id] = i
            new_data.append(f"{smi} {i}\n")
        with open(new_path, 'w') as f:
            for line in new_data:
                f.write(line)
        return new_path, mapping
    
    elif extension == 'sdf':
        suppl = Chem.SDMolSupplier(path, removeHs=False)
        mapping = {}
        with Chem.SDWriter(new_path) as w:
            for i, mol in enumerate(suppl):
                id = mol.GetProp("_Name").strip()
                mapping[id] = i
                mol.SetProp("_Name", str(i))
                w.write(mol)
        return new_path, mapping

    else:
        raise ValueError("The input file should be either smi or sdf")

def decode_ids(path: str, mapping: dict) -> str:
    '''For an SDF file, decode the ids using the mapping'''
    mapping = {v: k for k, v in mapping.items()}
    basename = os.path.basename(path)
    dir = os.path.dirname(os.path.abspath(path))
    extension = basename.split('.')[-1].strip()
    new_path = os.path.join(dir,
                            '_'.join(basename.split('.')[0].strip().split('_')[:-2]) + '_out.' + extension)
    
    suppl = Chem.SDMolSupplier(path, removeHs=False)
    with Chem.SDWriter(new_path) as w:
        for mol in suppl:
            name = mol.GetProp("_Name").strip()
            if '@taut' in name:
                components = name.split('@taut')
                new_name = mapping[int(components[0])] + '@taut' + ''.join(components[1:])
            else:
                new_name = mapping[int(name)]
            mol.SetProp("_Name", new_name)

            id = '_'.join(mol.GetProp("ID").strip().split('_')[1:])
            new_id = new_name + '_' + id
            mol.SetProp("ID", new_id)

            w.write(mol)
    return new_path


import os
import tempfile
import pickle
import pytest
import numpy as np
import torch
from rdkit import Chem
import Auto3D
from Auto3D.ASE.geometry import opt_geometry
from Auto3D.ASE.thermo import model_name2model_calculator, vib_hessian
from Auto3D.ASE.thermo import calc_thermo


folder = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

try:
    import torchani
    class userNNP1(torch.nn.Module):
        def __init__(self):
            super(userNNP1, self).__init__()
            """This is an example NNP model that can be used with Auto3D.
            You can initialize an NNP model however you want,
            just make sure that:
                - It contains the coord_pad and species_pad attributes 
                (These values will be used when processing the molecules in batch.)
                - The signature of the forward method is the same as below.
            """
            # Here I constructed an example NNP using ANI2x.
            # In your case, you can replace this with your own NNP model.
            self.model = torchani.models.ANI2x(periodic_table_index=True)

            self.coord_pad = 0  # int, the padding value for coordinates
            self.species_pad = -1  # int, the padding value for species.
            # self.state_dict = None

        def forward(self,
                    species: torch.Tensor,
                    coords: torch.Tensor,
                    charges: torch.Tensor) -> torch.Tensor:
            """
            Your NNP should take species, coords, and charges as input
            and return the energies of the molecules.

            species contains the atomic numbers of the atoms in the molecule: [B, N]
            where B is the batch size, N is the number of atoms in the largest molecule.
            
            coords contains the coordinates of the atoms in the molecule: [B, N, 3]
            where B is the batch size, N is the number of atoms in the largest molecule,
            and 3 represents the x, y, z coordinates.
            
            charges contains the molecular charges: [B]
            
            The forward function returns the energies of the molecules: [B],
            output energy unit: eV"""

            # an example for computing molecular energy, replace with your NNP model
            energies = self.model((species, coords)).energies * 27.211386245988
            return energies
    test_userNNP1 = True
except:
    test_userNNP1 = False

class userNNP2(torch.nn.Module):
    def __init__(self):
        super(userNNP2, self).__init__()
        """This is an example NNP model that can be used with Auto3D.
        You can initialize an NNP model however you want,
        just make sure that:
            - It contains the coord_pad and species_pad attributes 
            (These values will be used when processing the molecules in batch.)
            - The signature of the forward method is the same as below.
        """
        # Here I constructed an example NNP using AIMNet2.
        # In your case, you can replace this with your own NNP model.
        self.model = torch.jit.load(os.path.join(folder, 'src/Auto3D/models/aimnet2_wb97m-d3_0.jpt'))

        self.coord_pad = 0  # int, the padding value for coordinates
        self.species_pad = 0  # int, the padding value for species.
        # self.state_dict = None

    def forward(self,
                species: torch.Tensor,
                coords: torch.Tensor,
                charges: torch.Tensor) -> torch.Tensor:
        """
        Your NNP should take species, coords, and charges as input
        and return the energies of the molecules.

        species contains the atomic numbers of the atoms in the molecule: [B, N]
        where B is the batch size, N is the number of atoms in the largest molecule.
        
        coords contains the coordinates of the atoms in the molecule: [B, N, 3]
        where B is the batch size, N is the number of atoms in the largest molecule,
        and 3 represents the x, y, z coordinates.
        
        charges contains the molecular charges: [B]
        
        The forward function returns the energies of the molecules: [B],
        output energy unit: eV"""

        # an example for computing molecular energy, replace with your NNP model
        dct = dict(coord=coords, numbers=species, charge=charges)
        energies = self.model(dct)['energy']
        return energies


def test_calc_thermo_aimnet():
    #load wB97m-D4/Def2-TZVPP output file
    path = os.path.join(folder, "tests/files/cyclooctane.sdf")
    reference_G = -314.49236715
    reference_H = -314.45168666

    #compare Auto3D output with the above
    out = calc_thermo(path, "AIMNET", opt_tol=0.003)
    mol = next(Chem.SDMolSupplier(out, removeHs=False))

    G_out = float(mol.GetProp("G_hartree"))
    H_out = float(mol.GetProp("H_hartree"))
    assert(abs(reference_G - G_out) <= 0.02)
    assert(abs(reference_H - H_out) <= 0.02)
    try:
        os.remove(out)
    except:
        pass

def test_vib_hessian():
    path = os.path.join(folder, "tests/files/cyclooctane.sdf")
    mol = next(Chem.SDMolSupplier(path, removeHs=False))

    _, calculator = model_name2model_calculator('AIMNET')
    model_path = os.path.join(folder, "src/Auto3D/models/aimnet2_wb97m-d3_0.jpt")
    device = torch.device('cpu')
    model = torch.jit.load(model_path, map_location=device)

    hessian_vib = vib_hessian(mol, calculator, model)
    hessian_freq = hessian_vib.get_frequencies()

    ase_freq = pickle.load(open(os.path.join(folder, "tests/files/cyclooctane_ase_freq.pkl"), "rb"))
    mean_diff = np.mean(np.abs(hessian_freq[6:] - ase_freq[6:]))
    assert(mean_diff <= 10)  # 10 cm-1 error is acceptable

def test_opt_geometry1():
    path = os.path.join(folder, "tests/files/DA.sdf")
    out = opt_geometry(path, 'ANI2x', gpu_idx=0, opt_tol=0.1, opt_steps=5000)
    try:
        os.remove(out)
    except:
        pass

def test_opt_geometry2():
    path = os.path.join(folder, "tests/files/DA.sdf")
    out = opt_geometry(path, 'ANI2xt', gpu_idx=0, opt_tol=0.1, opt_steps=5000)
    try:
        os.remove(out)
    except:
        pass

def test_opt_geometry3():
    path = os.path.join(folder, "tests/files/DA.sdf")
    out = opt_geometry(path, 'AIMNET', gpu_idx=0, opt_tol=0.1, opt_steps=5000)
    try:
        os.remove(out)
    except:
        pass

@pytest.mark.skipif(not test_userNNP1, reason="TorchANI is not  installed.")
def test_opt_geometry4():
    path = os.path.join(folder, "tests/files/DA.sdf")
    with tempfile.TemporaryDirectory() as tmpdir:
        model_path = os.path.join(tmpdir, 'myNNP.pt')
        myNNP = userNNP1()
        myNNP_jit = torch.jit.script(myNNP)
        myNNP_jit.save(model_path)
    
        out = opt_geometry(path, model_path, gpu_idx=0, opt_tol=0.1, opt_steps=5000)
    try:
        os.remove(out)
    except:
        pass

def test_opt_geometry5():
    path = os.path.join(folder, "tests/files/DA.sdf")
    with tempfile.TemporaryDirectory() as tmpdir:
        model_path = os.path.join(tmpdir, 'myNNP.pt')
        myNNP = userNNP2()
        myNNP_jit = torch.jit.script(myNNP)
        myNNP_jit.save(model_path)
    
        out = opt_geometry(path, model_path, gpu_idx=0, opt_tol=0.1, opt_steps=5000)
    try:
        os.remove(out)
    except:
        pass


@pytest.mark.skipif(not test_userNNP1, reason="TorchANI is not  installed.")
def test_calc_thermo_userNNP1():
    #load wB97m-D4/Def2-TZVPP output file
    # Note that this is not the target DFT level for ANI2x
    # The purpose is just to verify the correctness of ani2x_jit
    path = os.path.join(folder, "tests/files/cyclooctane.sdf")

    #compute thermodynamic properties with ani2x_jit
    with tempfile.TemporaryDirectory() as tmpdir:
        model_path = os.path.join(tmpdir, 'myNNP.pt')
        myNNP = userNNP1()
        myNNP_jit = torch.jit.script(myNNP)
        myNNP_jit.save(model_path)
        out = calc_thermo(path, model_path, opt_tol=0.003)
    mol = next(Chem.SDMolSupplier(out, removeHs=False))

    G_out = float(mol.GetProp("G_hartree"))
    H_out = float(mol.GetProp("H_hartree"))

    # compute thermodynamic properties with ani2x
    out2 = calc_thermo(path, "ANI2x", opt_tol=0.003)
    mol2 = next(Chem.SDMolSupplier(out2, removeHs=False))
    G_out2 = float(mol2.GetProp("G_hartree"))
    H_out2 = float(mol2.GetProp("H_hartree"))

    assert(abs(G_out - G_out2) <= 0.02)
    assert(abs(H_out - H_out2) <= 0.02)
    try:
        os.remove(out)
    except:
        pass
    try:
        os.remove(out2)
    except:
        pass


def test_calc_thermo_userNNP2():
    #load wB97m-D4/Def2-TZVPP output file
    path = os.path.join(folder, "tests/files/cyclooctane.sdf")
    reference_G = -314.49236715
    reference_H = -314.45168666

    #compare Auto3D output with the above
    with tempfile.TemporaryDirectory() as tmpdir:
        model_path = os.path.join(tmpdir, 'myNNP.pt')
        myNNP = userNNP2()
        myNNP_jit = torch.jit.script(myNNP)
        myNNP_jit.save(model_path)
        out = calc_thermo(path, model_path, opt_tol=0.003)
    mol = next(Chem.SDMolSupplier(out, removeHs=False))

    G_out = float(mol.GetProp("G_hartree"))
    H_out = float(mol.GetProp("H_hartree"))
    assert(abs(reference_G - G_out) <= 0.02)
    assert(abs(reference_H - H_out) <= 0.02)
    try:
        os.remove(out)
    except:
        pass


if __name__ == "__main__":
    print()
    # test_calc_thermo_aimnet()
    test_calc_thermo_userNNP1()
    test_calc_thermo_userNNP2()

    # from Auto3D.ASE.thermo import mol2aimnet_input

    # device = torch.device('cpu')
    # path = os.path.join(folder, 'tests/files/cyclooctane.sdf')
    # e_ref = -314.689736079491
    # supp = Chem.SDMolSupplier(path, removeHs=False)
    # print(f'Number of conformers: {len(supp)}')
    # mol = supp[0]
    

    # # original ani2x
    # ani2x = torchani.models.ANI2x()
    # dct = mol2aimnet_input(mol, device)
    # dct['coord'].requires_grad = True
    # out = aimnet2(dct)
    # e = out['energy']
    # f = - torch.autograd.grad(e, dct['coord'])[0]
    # print(e)
    # print(f)


    # myNNP2
    # myNNP = userNNP2()
    # myNNP_jit = torch.jit.script(myNNP)
    # myNNP_jit.save('/home/jack/Auto3D_pkg/example/myNNP2.pt')

    # myNNP = torch.jit.load('/home/jack/Auto3D_pkg/example/myNNP2.pt', map_location=device).double()
    
    # my_e = myNNP(dct['numbers'], dct['coord'], dct['charge'])
    # print(my_e)

    # my_f = - torch.autograd.grad(my_e, dct['coord'])[0]
    # print(my_f)

    # f_diff = torch.sum(torch.abs(f - my_f))
    # print(f_diff)


#!/usr/bin/env python
"""
Generating low-energy conformers from SMILES.
"""
import logging
# import argparse
import os
import shutil
import sys
import time
from datetime import datetime
import torch
import torch.nn as nn
import math
import psutil, tarfile
import glob
import pandas as pd
import multiprocessing as mp
import tempfile
from logging.handlers import QueueHandler
from typing import List, Optional, Union
from rdkit import Chem
import Auto3D
from Auto3D.isomer_engine import rd_isomer, tautomer_engine
from Auto3D.isomer_engine import rd_isomer_sdf
from Auto3D.isomer_engine import oe_isomer
from Auto3D.ranking import ranking
from Auto3D.batch_opt.batchopt import optimizing
from Auto3D.utils import housekeeping
from Auto3D.utils import check_input
from Auto3D.utils import hash_taut_smi,  my_name_space
from Auto3D.utils import create_chunk_meta_names
from Auto3D.utils import reorder_sdf
from Auto3D.utils_file import SDF2chunks
from Auto3D.utils_file import smiles2smi
from Auto3D.utils_file import encode_ids, decode_ids
from send2trash import send2trash
try:
    mp.set_start_method('spawn')
except:
    pass

torch.backends.cuda.matmul.allow_tf32 = False
torch.backends.cudnn.allow_tf32 = False

def isomer_wraper(chunk_info, args, queue, logging_queue):
    """
    chunk_info: (path, dir) tuple for the chunk
    args: auto3D arguments
    queue: mp.queue
    logging_queue
    """
    #prepare logging
    logger = logging.getLogger("auto3d")
    logger.addHandler(QueueHandler(logging_queue))
    logger.setLevel(logging.INFO)

    for i, path_dir in enumerate(chunk_info):
        print(f"\n\nIsomer generation for job{i+1}", flush=True)
        logger.info(f"\n\nIsomer generation for job{i+1}")
        path, dir = path_dir
        meta = create_chunk_meta_names(path, dir)

        # Tautomer enumeratioin
        if args.enumerate_tautomer:
            output_taut = meta["output_taut"]
            taut_mode = args.tauto_engine
            print("Enumerating tautomers for the input...", end='')
            logger.info("Enumerating tautomers for the input...")
            taut_engine = tautomer_engine(taut_mode, path, output_taut, args.pKaNorm)
            taut_engine.run()
            hash_taut_smi(output_taut, output_taut)
            path = output_taut
            print(f"Tautomers are saved in {output_taut}", flush=True)
            logger.info(f"Tautomers are saved in {output_taut}")

        smiles_enumerated = meta["smiles_enumerated"]
        smiles_reduced = meta["smiles_reduced"]
        smiles_hashed = meta["smiles_hashed"]
        enumerated_sdf = meta["enumerated_sdf"]
        max_confs = args.max_confs
        duplicate_threshold = args.threshold
        mpi_np = args.mpi_np
        enumerate_isomer = args.enumerate_isomer
        isomer_program = args.isomer_engine
        # Isomer enumeration step
        if isomer_program == 'omega':
            mode_oe = args.mode_oe
            oe_isomer(mode_oe, path, smiles_enumerated, smiles_reduced, smiles_hashed,
                    enumerated_sdf, max_confs, duplicate_threshold, enumerate_isomer)
        elif isomer_program == 'rdkit':
            if args.input_format == 'smi':
                engine = rd_isomer(path, smiles_enumerated, smiles_reduced, smiles_hashed, 
                                enumerated_sdf, dir, max_confs, duplicate_threshold, mpi_np, enumerate_isomer)
                engine.run()
            elif args.input_format == 'sdf':
                engine = rd_isomer_sdf(path, enumerated_sdf, max_confs, duplicate_threshold, mpi_np)
                engine.run()
        else: 
            raise ValueError('The isomer enumeration engine must be "omega" or "rdkit", '
                            f'but {args.isomer_engine} was parsed. '
                            'You can set the parameter by appending the following:'
                            '--isomer_engine=rdkit')

        queue.put((enumerated_sdf, path, dir, i+1))
    if isinstance(args.gpu_idx, int) or len(args.gpu_idx) == 1:
        queue.put("Done")
    else:
        for _ in range(len(args.gpu_idx)):
            queue.put("Done")


def optim_rank_wrapper(args, queue, logging_queue, gpu_idx:int) -> List[Chem.Mol]:
    #prepare logging
    logger = logging.getLogger("auto3d")
    logger.addHandler(QueueHandler(logging_queue))
    logger.setLevel(logging.INFO)

    conformers = []
    while True:
        sdf_path_dir_job = queue.get()
        if sdf_path_dir_job == "Done":
            break
        enumerated_sdf, path, dir, job = sdf_path_dir_job
        print(f"\n\nOptimizing on job{job}", flush=True)
        logger.info(f"\n\nOptimizing on job{job}")
        meta = create_chunk_meta_names(path, dir)

        # Optimizing step
        opt_steps = args.opt_steps
        opt_tol = args.convergence_threshold
        patience = args.patience
        batchsize_atoms = args.batchsize_atoms
        config = {"opt_steps": opt_steps, "opttol": opt_tol, "patience": patience, "batchsize_atoms": batchsize_atoms}
        optimized_og = meta["optimized_og"]
        optimizing_engine = args.optimizing_engine
        if args.use_gpu:
            device = torch.device(f"cuda:{gpu_idx}")
        else:
            device = torch.device("cpu")
        optimizer = optimizing(enumerated_sdf, optimized_og,
                               optimizing_engine, device, config)
        optimizer.run()

        # Ranking step
        output = meta["output"]
        duplicate_threshold = args.threshold
        k = args.k
        window = args.window
        rank_engine = ranking(optimized_og,
                              output, duplicate_threshold, k=k, window=window)
        conformers.append(rank_engine.run())

        # Housekeeping
        housekeeping_folder = meta["housekeeping_folder"]
        os.mkdir(housekeeping_folder)
        housekeeping(dir, housekeeping_folder, output)
        #Conpress verbose folder
        housekeeping_folder_gz = housekeeping_folder + ".tar.gz"
        with tarfile.open(housekeeping_folder_gz, "w:gz") as tar:
            tar.add(housekeeping_folder, arcname=os.path.basename(housekeeping_folder))
        shutil.rmtree(housekeeping_folder)
        if not args.verbose:
            try:  #Clusters does not support send2trash
                send2trash(housekeeping_folder_gz)
            except:
                os.remove(housekeeping_folder_gz)
    return conformers

def options(path: Optional[str]=None, k=False, window=False, verbose=False, job_name="",
    enumerate_tautomer=False, tauto_engine="rdkit", pKaNorm=True,
    isomer_engine="rdkit", enumerate_isomer=True, mode_oe="classic", mpi_np=4, max_confs=None,
    use_gpu=True, gpu_idx: Union[int, List[int]]=0, capacity=42, optimizing_engine="AIMNET", patience=1000,
    opt_steps=5000, convergence_threshold=0.003, threshold=0.3, memory=None, batchsize_atoms=1024):
    """
    Generating arguments for the Auto3D ``main`` function.

    :param path: A input.smi containing SMILES and IDs. Examples are listed in the example/files folder
    :type path: str, optional
    :param k: Outputs the top-k structures for each SMILES, defaults to False
    :type k: bool, optional
    :param window: Outputs the structures whose energies are within x (kcal/mol) from the lowest energy conformer, defaults to False
    :type window: bool, optional
    :param verbose: When True, save all meta data while running, defaults to False
    :type verbose: bool, optional
    :param job_name: A folder name to save all meta data, defaults to ""
    :type job_name: str, optional
    :param enumerate_tautomer: When True, enumerate tautomers for the input, defaults to False
    :type enumerate_tautomer: bool, optional
    :param tauto_engine: Programs to enumerate tautomers, either 'rdkit' or 'oechem', defaults to "rdkit"
    :type tauto_engine: str, optional
    :param pKaNorm: When True, the ionization state of each tautomer will be assigned to a predominant state at ~7.4 (Only works when tauto_engine='oechem'), defaults to True
    :type pKaNorm: bool, optional
    :param isomer_engine: The program for generating 3D isomers for each SMILES. This parameter is either rdkit or omega, defaults to "rdkit"
    :type isomer_engine: str, optional
    :param enumerate_isomer: When True, cis/trans and r/s isomers are enumerated, defaults to True
    :type enumerate_isomer: bool, optional
    :param mode_oe: The mode that omega program will take. It can be either 'classic', 'macrocycle', 'dense', 'pose', 'rocs' or 'fast_rocs'. By default, the 'classic' mode is used. For detailed information about each mode, see https://docs.eyesopen.com/applications/omega/omega/omega_overview.html, defaults to "classic"
    :type mode_oe: str, optional
    :param mpi_np: Number of CPU cores for the isomer generation engine, defaults to 4
    :type mpi_np: int, optional
    :param max_confs: Maximum number of isomers for each SMILES. Default is None, and Auto3D will uses a dynamic conformer number for each SMILES. The number of conformer for each SMILES is the number of heavey atoms in the SMILES minus 1, defaults to None
    :type max_confs: int, optional
    :param use_gpu: If True, the program will use GPU when available, defaults to True
    :type use_gpu: bool, optional
    :param gpu_idx: GPU index. It only works when --use_gpu=True, defaults to 0
    :type gpu_idx: int or list of int, optional
    :param capacity: Number of SMILES that the model will handle for 1 G memory, defaults to 42
    :type capacity: int, optional
    :param optimizing_engine: Choose either 'ANI2x', 'ANI2xt', 'AIMNET' or a path to a custom NNP for energy calculation and geometry optimization, defaults to "AIMNET"
    :type optimizing_engine: str, optional
    :param patience: If the force does not decrease for a continuous patience steps, the conformer will drop out of the optimization loop, defaults to 1000
    :type patience: int, optional
    :param opt_steps: Maximum optimization steps for each structure, defaults to 5000
    :type opt_steps: int, optional
    :param convergence_threshold: Optimization is considered as converged if maximum force is below this threshold, defaults to 0.003
    :type convergence_threshold: float, optional
    :param threshold: If the RMSD between two conformers are within threhold, they are considered as duplicates. One of them will be removed, defaults to 0.3
    :type threshold: float, optional
    :param memory: The RAM size assigned to Auto3D (unit GB), defaults to None
    :type memory: int, optional
    :param batchsize_atoms: The number of atoms in 1 optimization batch for 1GB, defaults to 1024
    :type batchsize_atoms: int, optional
    """
    d = {}
    args = my_name_space(d)
    args['path'] = path
    args['k'] = k
    args['window'] = window
    args['verbose'] = verbose
    args['job_name'] = job_name
    args["enumerate_tautomer"] = enumerate_tautomer
    args["tauto_engine"] = tauto_engine.lower()
    args["pKaNorm"] = pKaNorm
    args["isomer_engine"] = isomer_engine.lower()
    args["enumerate_isomer"] = enumerate_isomer
    args["mode_oe"] = mode_oe.lower()
    args["mpi_np"] = mpi_np
    args["max_confs"] = max_confs
    args["use_gpu"] = use_gpu
    args["capacity"] = capacity
    args["gpu_idx"] = gpu_idx
    args["optimizing_engine"] = optimizing_engine
    args["patience"] = patience
    args["opt_steps"] = opt_steps
    args["convergence_threshold"] = convergence_threshold
    args["threshold"] = threshold
    args["memory"] = memory
    args["batchsize_atoms"] = batchsize_atoms
    return args


def logger_process(queue, logging_path):
    """A child process for logging all information from other processes"""
    logger = logging.getLogger("auto3d")
    logger.addHandler(logging.FileHandler(logging_path))
    logger.setLevel(logging.INFO)
    while True:
        message = queue.get()
        if message is None:
            break
        logger.handle(message)


def main(args:dict):
    """Take the arguments from the ``options`` function and run Auto3D."""
    # Ensure spawn method is used
    try:
        mp.set_start_method('fork')
    except RuntimeError:
        pass

    chunk_line = mp.Manager().Queue()   #A queue managing two wrappers
    start = time.time()
    # job_name = datetime.now().strftime("%Y%m%d-%H%M%S-%f")  #adds microsecond in the end

    if args.path is None:
        sys.exit("Please specify the input file path.")
    path0, mapping = encode_ids(args.path)
    input_format = os.path.splitext(path0)[1][1:]
    if (input_format != "smi") and (input_format != "sdf"):
        sys.exit("Input file type is not supported. Only .smi and .sdf are supported. But the input file is " + input_format + ".")
    args['input_format'] = input_format
    k = args.k
    window = args.window
    if (not k) and (not window):
        sys.exit("Either k or window needs to be specified. "
                "Usually, setting '--k=1' satisfies most needs.")
    if args.job_name == "":
        args.job_name = datetime.now().strftime("%Y%m%d-%H%M%S-%f")  #adds microsecond in the end
    job_name = args.job_name

    # initialiazation
    basename = os.path.basename(path0)
    dir = os.path.dirname(os.path.abspath(path0))
    # job_name = job_name + "_" + basename.split('.')[0].strip()
    job_name =  basename.split('.')[0].strip()[:-8] + '_' + job_name  #remove '_encoded'
    job_name = os.path.join(dir, job_name)
    os.mkdir(job_name)

    # initialize the logging process
    logging_path = os.path.join(job_name, "Auto3D.log")
    logging_queue = mp.Manager().Queue(999)
    logger_p = mp.Process(target=logger_process, args=(logging_queue, logging_path), daemon=True)
    logger_p.start()

    # logger in the main process
    logger = logging.getLogger("auto3d")
    logger.addHandler(QueueHandler(logging_queue))
    logger.setLevel(logging.INFO)
    logger.info(f"""
         _              _             _____   ____  
        / \     _   _  | |_    ___   |___ /  |  _ \ 
       / _ \   | | | | | __|  / _ \    |_ \  | | | |
      / ___ \  | |_| | | |_  | (_) |  ___) | | |_| |
     /_/   \_\  \__,_|  \__|  \___/  |____/  |____/  {Auto3D.__version__}
              // Generating low-energy 3D structures                                      
    """)

    logger.info("================================================================================")
    logger.info("                               INPUT PARAMETERS")
    logger.info("================================================================================")
    for key, val in args.items():
        line = str(key) + ": " + str(val)
        logger.info(line)

    logger.info("================================================================================")
    logger.info("                               RUNNING PROCESS")
    logger.info("================================================================================")

    check_input(args)
    # Devide jobs based on memory
    smiles_per_G = args.capacity  #Allow 40 SMILES per GB memory
    num_jobs = 1
    if args.memory is not None:
        t = int(args.memory)
    else:
        if args.use_gpu:
            if isinstance(args.gpu_idx, int):
                gpu_idx = int(args.gpu_idx)
            else:
                gpu_idx = args.gpu_idx[0]
                num_jobs = len(args.gpu_idx)
            t = int(math.ceil(torch.cuda.get_device_properties(gpu_idx).total_memory/(1024**3)))
        else:
            t = int(psutil.virtual_memory().total/(1024**3))
    chunk_size = t * smiles_per_G
    #batchsize_atoms based on GPU memory
    args.batchsize_atoms = args.batchsize_atoms * t

    #Get indexes for each chunk
    if input_format == "smi":
        df = pd.read_csv(path0, sep='\s+', header=None)
    elif input_format == "sdf":
        df = SDF2chunks(path0)
    data_size = len(df)
    num_chunks = max(int(data_size // chunk_size + 1), num_jobs)
    print(f"The available memory is {t} GB.", flush=True)
    print(f"The task will be divided into {num_chunks} jobs.", flush=True)
    logger.info(f"The available memory is {t} GB.")
    logger.info(f"The task will be divided into {num_chunks} jobs.")
    chunk_idxes = [[] for _ in range(num_chunks)]
    for i in range(num_chunks):
        idx = i
        while idx < data_size:
            chunk_idxes[i].append(idx)
            idx += num_chunks

    #Save each chunk as smi
    chunk_info = []
    basename = os.path.basename(path0).split(".")[0].strip()
    if input_format == "smi":
        for i in range(num_chunks):
            dir = os.path.join(job_name, f"job{i+1}")
            os.mkdir(dir)
            new_basename = basename + "_" + str(i+1) + ".smi"
            new_name = os.path.join(dir, new_basename)
            df_i = df.iloc[chunk_idxes[i], :]
            df_i.to_csv(new_name, header=None, index=None, sep=" ")
            path = new_name

            print(f"Job{i+1}, number of inputs: {len(df_i)}", flush=True)
            logger.info(f"Job{i+1}, number of inputs: {len(df_i)}")
            chunk_info.append((path, dir))
    elif input_format == "sdf":
        for i in range(num_chunks):
            dir = os.path.join(job_name, f"job{i+1}")
            os.mkdir(dir)
            new_basename = basename + "_" + str(i+1) + ".sdf"
            new_name = os.path.join(dir, new_basename)
            chunks_i = [df[j] for j in chunk_idxes[i]]
            with open(new_name, "w") as f:
                for chunk in chunks_i:
                    for line in chunk:
                        f.write(line)
            path = new_name    
            print(f"Job{i+1}, number of inputs: {len(chunks_i)}", flush=True)
            logger.info(f"Job{i+1}, number of inputs: {len(chunks_i)}")
            chunk_info.append((path, dir))

    p1 = mp.Process(target=isomer_wraper, args=(chunk_info, args, chunk_line, logging_queue,))
    p2s =  []
    if isinstance(args.gpu_idx, int):
        p2s.append(mp.Process(target=optim_rank_wrapper, args=(args, chunk_line, logging_queue, args.gpu_idx)))
    else:
        for idx in args.gpu_idx:
            p2s.append(mp.Process(target=optim_rank_wrapper, args=(args, chunk_line, logging_queue, idx)))
    p1.start()
    for p2 in p2s:
        p2.start()
    p1.join()
    for p2 in p2s:
        p2.join()

    #Combine jobs into a single sdf
    data = []
    paths = os.path.join(job_name, "job*/*_3d.sdf")
    files = glob.glob(paths)
    if len(files) == 0:
        msg = """The optimization engine did not run, or no 3D structure converged.
                 The reason might be one of the following: 
                 1. Allocated memory is not enough;
                 2. The input SMILES encodes invalid chemical structures;
                 3. Patience is too small"""
        sys.exit(msg)
    for file in files:
        with open(file, "r") as f:
            data_i = f.readlines()
        data += data_i
    combined_basename = basename + "_out.sdf"
    path_combined = os.path.join(job_name, combined_basename)
    with open(path_combined, "w+") as f:
        for line in data:
            f.write(line)

    # Program ends
    end = time.time()
    print("Energy unit: Hartree if implicit.", flush=True)
    logger.info("Energy unit: Hartree if implicit.")
    running_time_m = int((end - start)/60)
    if running_time_m <= 60:
        print(f'Program running time: {running_time_m + 1} minute(s)', flush=True)
        logger.info(f'Program running time: {running_time_m + 1} minute(s)')
    else:
        running_time_h = running_time_m // 60
        remaining_minutes = running_time_m - running_time_h*60
        print(f'Program running time: {running_time_h} hour(s) and {remaining_minutes} minute(s)', flush=True)
        logger.info(f'Program running time: {running_time_h} hour(s) and {remaining_minutes} minute(s)')
    reorder_sdf(path_combined, path0)
    path_output = decode_ids(path_combined, mapping)
    os.remove(path0)
    os.remove(path_combined)
    print(f"Output path: {path_output}", flush=True)
    logger.info(f"Output path: {path_output}")
    logging_queue.put(None)
    time.sleep(3)  #wait the daemon process for 3 seconds
    return path_output

def smiles2mols(smiles: List[str], args:dict) -> List[Chem.Mol]:
    """
    A handy tool for finding the low-energy conformers for a list of SMILES.
    Compared with the ``main`` function, it sacrifices efficiency for convenience.
    because ``smiles2mols`` uses only 1 process. 
    Both the input and output are returned as variables within Python.

    It's recommended only when the number of SMILES is less than 150;
    Otherwise using the main function will be faster.

    :param smiles: A list of SMILES strings for which to find low-energy conformers.
    :type smiles: List[str]
    :param args: A dictionary of arguments as returned by the ``option`` function.
    :type args: dict
    :return: A list of RDKit Mol objects representing the low-energy conformers of the input SMILES.
    :rtype: List[Chem.Mol]
    """
    with tempfile.TemporaryDirectory() as tmpdirname:
        basename = 'smiles.smi'
        path0 = os.path.join(tmpdirname, basename)
        smiles2smi(smiles, path0)  # save all SMILES into a smi file
        args['path'] = path0
        k = args.k
        window = args.window
        if (not k) and (not window):
            sys.exit("Either k or window needs to be specified. "
                    "Usually, setting '--k=1' satisfies most needs.")
        args.input_format = 'smi'
        check_input(args)

        # smi to sdf
        meta = create_chunk_meta_names(path0, tmpdirname)
        isomer_engine = rd_isomer(path0, meta["smiles_enumerated"],
                                  meta["smiles_reduced"], meta["smiles_hashed"], 
                                  meta["enumerated_sdf"], tmpdirname,
                                  args.max_confs, 0.03,
                                  args.mpi_np, args.enumerate_isomer)
        isomer_engine.run()

        # optimize conformers
        if args.use_gpu:
            if isinstance(args.gpu_idx, int):
                idx = args.gpu_idx
            else:
                idx = args.gpu_idx[0]
            device = torch.device(f"cuda:{idx}")
        else:
            device = torch.device("cpu")
        config = {"opt_steps": args.opt_steps, "opttol": args.convergence_threshold,
                  "patience": args.patience, "batchsize_atoms": args.batchsize_atoms}
        opt_engine = optimizing(meta["enumerated_sdf"], meta["optimized_og"],
                                args.optimizing_engine, device, config)
        opt_engine.run()

        # Ranking step
        rank_engine = ranking(meta["optimized_og"], meta["output"],
                              args.threshold, k=k, window=window)
        _ = rank_engine.run()
        conformers = reorder_sdf(meta["output"], path0)

        print("Energy unit: Hartree if implicit.", flush=True)
    return conformers
